# coding: utf-8

"""
    Graphsignal API

    API for uploading and querying spans, errors, metrics, and logs.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from datetime import datetime
from pydantic import Field, StrictInt, StrictStr
from typing import Any, Dict, List, Optional
from typing_extensions import Annotated
from graphsignal.client.models.error import Error
from graphsignal.client.models.error_query_result import ErrorQueryResult
from graphsignal.client.models.log_entry import LogEntry
from graphsignal.client.models.metric import Metric
from graphsignal.client.models.profile import Profile
from graphsignal.client.models.profile_query_result import ProfileQueryResult
from graphsignal.client.models.span import Span
from graphsignal.client.models.span_query_result import SpanQueryResult

from graphsignal.client.api_client import ApiClient, RequestSerialized
from graphsignal.client.api_response import ApiResponse
from graphsignal.client.rest import RESTResponseType


class DefaultApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    def query_errors(
        self,
        name: Annotated[Optional[StrictStr], Field(description="Filter errors by name. Provide error name.")] = None,
        tags: Annotated[Optional[StrictStr], Field(description="Filter errors by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        level: Annotated[Optional[StrictStr], Field(description="Filter errors by minimum level (debug, info, warning, error, critical). Only errors with this level or higher will be returned.")] = None,
        text: Annotated[Optional[StrictStr], Field(description="Filter errors by text. Provide text to search for in error messages.")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering errors, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering errors, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned errors. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of errors to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ErrorQueryResult:
        """Query errors

        Query errors by name, tags, text, and time range. Returns a list of errors that match the query parameters.

        :param name: Filter errors by name. Provide error name.
        :type name: str
        :param tags: Filter errors by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param level: Filter errors by minimum level (debug, info, warning, error, critical). Only errors with this level or higher will be returned.
        :type level: str
        :param text: Filter errors by text. Provide text to search for in error messages.
        :type text: str
        :param start_time: Start time for filtering errors, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering errors, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned errors. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of errors to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_errors_serialize(
            name=name,
            tags=tags,
            level=level,
            text=text,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ErrorQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def query_errors_with_http_info(
        self,
        name: Annotated[Optional[StrictStr], Field(description="Filter errors by name. Provide error name.")] = None,
        tags: Annotated[Optional[StrictStr], Field(description="Filter errors by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        level: Annotated[Optional[StrictStr], Field(description="Filter errors by minimum level (debug, info, warning, error, critical). Only errors with this level or higher will be returned.")] = None,
        text: Annotated[Optional[StrictStr], Field(description="Filter errors by text. Provide text to search for in error messages.")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering errors, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering errors, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned errors. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of errors to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[ErrorQueryResult]:
        """Query errors

        Query errors by name, tags, text, and time range. Returns a list of errors that match the query parameters.

        :param name: Filter errors by name. Provide error name.
        :type name: str
        :param tags: Filter errors by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param level: Filter errors by minimum level (debug, info, warning, error, critical). Only errors with this level or higher will be returned.
        :type level: str
        :param text: Filter errors by text. Provide text to search for in error messages.
        :type text: str
        :param start_time: Start time for filtering errors, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering errors, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned errors. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of errors to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_errors_serialize(
            name=name,
            tags=tags,
            level=level,
            text=text,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ErrorQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def query_errors_without_preload_content(
        self,
        name: Annotated[Optional[StrictStr], Field(description="Filter errors by name. Provide error name.")] = None,
        tags: Annotated[Optional[StrictStr], Field(description="Filter errors by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        level: Annotated[Optional[StrictStr], Field(description="Filter errors by minimum level (debug, info, warning, error, critical). Only errors with this level or higher will be returned.")] = None,
        text: Annotated[Optional[StrictStr], Field(description="Filter errors by text. Provide text to search for in error messages.")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering errors, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering errors, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned errors. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of errors to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Query errors

        Query errors by name, tags, text, and time range. Returns a list of errors that match the query parameters.

        :param name: Filter errors by name. Provide error name.
        :type name: str
        :param tags: Filter errors by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param level: Filter errors by minimum level (debug, info, warning, error, critical). Only errors with this level or higher will be returned.
        :type level: str
        :param text: Filter errors by text. Provide text to search for in error messages.
        :type text: str
        :param start_time: Start time for filtering errors, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering errors, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned errors. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of errors to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_errors_serialize(
            name=name,
            tags=tags,
            level=level,
            text=text,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ErrorQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _query_errors_serialize(
        self,
        name,
        tags,
        level,
        text,
        start_time,
        end_time,
        limit,
        offset,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if name is not None:
            
            _query_params.append(('name', name))
            
        if tags is not None:
            
            _query_params.append(('tags', tags))
            
        if level is not None:
            
            _query_params.append(('level', level))
            
        if text is not None:
            
            _query_params.append(('text', text))
            
        if start_time is not None:
            if isinstance(start_time, datetime):
                _query_params.append(
                    (
                        'start_time',
                        start_time.strftime(
                            self.api_client.configuration.datetime_format
                        )
                    )
                )
            else:
                _query_params.append(('start_time', start_time))
            
        if end_time is not None:
            if isinstance(end_time, datetime):
                _query_params.append(
                    (
                        'end_time',
                        end_time.strftime(
                            self.api_client.configuration.datetime_format
                        )
                    )
                )
            else:
                _query_params.append(('end_time', end_time))
            
        if limit is not None:
            
            _query_params.append(('limit', limit))
            
        if offset is not None:
            
            _query_params.append(('offset', offset))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/api/v1/errors/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def query_profiles(
        self,
        name: Annotated[Optional[StrictStr], Field(description="Filter profiles by name. Provide profile name.")] = None,
        tags: Annotated[Optional[StrictStr], Field(description="Filter profiles by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        format: Annotated[Optional[StrictStr], Field(description="Filter profiles by format (e.g., chrome-trace, pprof).")] = None,
        text: Annotated[Optional[StrictStr], Field(description="Filter profiles by text. Provide text to search for in profile names or content.")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering profiles, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering profiles, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned profiles. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of profiles to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ProfileQueryResult:
        """Query profiles

        Query profiles by name, tags, format, text, and time range. Returns a list of profiles that match the query parameters.

        :param name: Filter profiles by name. Provide profile name.
        :type name: str
        :param tags: Filter profiles by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param format: Filter profiles by format (e.g., chrome-trace, pprof).
        :type format: str
        :param text: Filter profiles by text. Provide text to search for in profile names or content.
        :type text: str
        :param start_time: Start time for filtering profiles, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering profiles, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned profiles. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of profiles to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_profiles_serialize(
            name=name,
            tags=tags,
            format=format,
            text=text,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ProfileQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def query_profiles_with_http_info(
        self,
        name: Annotated[Optional[StrictStr], Field(description="Filter profiles by name. Provide profile name.")] = None,
        tags: Annotated[Optional[StrictStr], Field(description="Filter profiles by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        format: Annotated[Optional[StrictStr], Field(description="Filter profiles by format (e.g., chrome-trace, pprof).")] = None,
        text: Annotated[Optional[StrictStr], Field(description="Filter profiles by text. Provide text to search for in profile names or content.")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering profiles, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering profiles, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned profiles. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of profiles to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[ProfileQueryResult]:
        """Query profiles

        Query profiles by name, tags, format, text, and time range. Returns a list of profiles that match the query parameters.

        :param name: Filter profiles by name. Provide profile name.
        :type name: str
        :param tags: Filter profiles by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param format: Filter profiles by format (e.g., chrome-trace, pprof).
        :type format: str
        :param text: Filter profiles by text. Provide text to search for in profile names or content.
        :type text: str
        :param start_time: Start time for filtering profiles, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering profiles, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned profiles. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of profiles to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_profiles_serialize(
            name=name,
            tags=tags,
            format=format,
            text=text,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ProfileQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def query_profiles_without_preload_content(
        self,
        name: Annotated[Optional[StrictStr], Field(description="Filter profiles by name. Provide profile name.")] = None,
        tags: Annotated[Optional[StrictStr], Field(description="Filter profiles by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        format: Annotated[Optional[StrictStr], Field(description="Filter profiles by format (e.g., chrome-trace, pprof).")] = None,
        text: Annotated[Optional[StrictStr], Field(description="Filter profiles by text. Provide text to search for in profile names or content.")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering profiles, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering profiles, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned profiles. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of profiles to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Query profiles

        Query profiles by name, tags, format, text, and time range. Returns a list of profiles that match the query parameters.

        :param name: Filter profiles by name. Provide profile name.
        :type name: str
        :param tags: Filter profiles by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param format: Filter profiles by format (e.g., chrome-trace, pprof).
        :type format: str
        :param text: Filter profiles by text. Provide text to search for in profile names or content.
        :type text: str
        :param start_time: Start time for filtering profiles, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering profiles, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned profiles. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of profiles to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_profiles_serialize(
            name=name,
            tags=tags,
            format=format,
            text=text,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ProfileQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _query_profiles_serialize(
        self,
        name,
        tags,
        format,
        text,
        start_time,
        end_time,
        limit,
        offset,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if name is not None:
            
            _query_params.append(('name', name))
            
        if tags is not None:
            
            _query_params.append(('tags', tags))
            
        if format is not None:
            
            _query_params.append(('format', format))
            
        if text is not None:
            
            _query_params.append(('text', text))
            
        if start_time is not None:
            if isinstance(start_time, datetime):
                _query_params.append(
                    (
                        'start_time',
                        start_time.strftime(
                            self.api_client.configuration.datetime_format
                        )
                    )
                )
            else:
                _query_params.append(('start_time', start_time))
            
        if end_time is not None:
            if isinstance(end_time, datetime):
                _query_params.append(
                    (
                        'end_time',
                        end_time.strftime(
                            self.api_client.configuration.datetime_format
                        )
                    )
                )
            else:
                _query_params.append(('end_time', end_time))
            
        if limit is not None:
            
            _query_params.append(('limit', limit))
            
        if offset is not None:
            
            _query_params.append(('offset', offset))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/api/v1/profiles/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def query_spans(
        self,
        tags: Annotated[Optional[StrictStr], Field(description="Filter spans by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering spans, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering spans, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned spans. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of spans to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> SpanQueryResult:
        """Query spans

        Query spans by tags, text, and time range. Returns a list of spans that match the query parameters.

        :param tags: Filter spans by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param start_time: Start time for filtering spans, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering spans, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned spans. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of spans to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_spans_serialize(
            tags=tags,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SpanQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def query_spans_with_http_info(
        self,
        tags: Annotated[Optional[StrictStr], Field(description="Filter spans by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering spans, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering spans, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned spans. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of spans to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[SpanQueryResult]:
        """Query spans

        Query spans by tags, text, and time range. Returns a list of spans that match the query parameters.

        :param tags: Filter spans by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param start_time: Start time for filtering spans, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering spans, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned spans. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of spans to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_spans_serialize(
            tags=tags,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SpanQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def query_spans_without_preload_content(
        self,
        tags: Annotated[Optional[StrictStr], Field(description="Filter spans by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').")] = None,
        start_time: Annotated[Optional[datetime], Field(description="Start time for filtering spans, in ISO format.")] = None,
        end_time: Annotated[Optional[datetime], Field(description="End time for filtering spans, in ISO format.")] = None,
        limit: Annotated[Optional[StrictInt], Field(description="Limit the number of returned spans. Default is 100. Max is 1000.")] = None,
        offset: Annotated[Optional[StrictInt], Field(description="The number of spans to skip before starting to collect the result set.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Query spans

        Query spans by tags, text, and time range. Returns a list of spans that match the query parameters.

        :param tags: Filter spans by tags. Provide tags as a semicolon-separated list of key-value pairs (e.g., 'key1:value1;key2:value2').
        :type tags: str
        :param start_time: Start time for filtering spans, in ISO format.
        :type start_time: datetime
        :param end_time: End time for filtering spans, in ISO format.
        :type end_time: datetime
        :param limit: Limit the number of returned spans. Default is 100. Max is 1000.
        :type limit: int
        :param offset: The number of spans to skip before starting to collect the result set.
        :type offset: int
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._query_spans_serialize(
            tags=tags,
            start_time=start_time,
            end_time=end_time,
            limit=limit,
            offset=offset,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SpanQueryResult",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _query_spans_serialize(
        self,
        tags,
        start_time,
        end_time,
        limit,
        offset,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if tags is not None:
            
            _query_params.append(('tags', tags))
            
        if start_time is not None:
            if isinstance(start_time, datetime):
                _query_params.append(
                    (
                        'start_time',
                        start_time.strftime(
                            self.api_client.configuration.datetime_format
                        )
                    )
                )
            else:
                _query_params.append(('start_time', start_time))
            
        if end_time is not None:
            if isinstance(end_time, datetime):
                _query_params.append(
                    (
                        'end_time',
                        end_time.strftime(
                            self.api_client.configuration.datetime_format
                        )
                    )
                )
            else:
                _query_params.append(('end_time', end_time))
            
        if limit is not None:
            
            _query_params.append(('limit', limit))
            
        if offset is not None:
            
            _query_params.append(('offset', offset))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/api/v1/spans/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def upload_errors(
        self,
        error: List[Error],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> object:
        """Upload errors

        Upload errors

        :param error: (required)
        :type error: List[Error]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_errors_serialize(
            error=error,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def upload_errors_with_http_info(
        self,
        error: List[Error],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[object]:
        """Upload errors

        Upload errors

        :param error: (required)
        :type error: List[Error]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_errors_serialize(
            error=error,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def upload_errors_without_preload_content(
        self,
        error: List[Error],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Upload errors

        Upload errors

        :param error: (required)
        :type error: List[Error]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_errors_serialize(
            error=error,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _upload_errors_serialize(
        self,
        error,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'Error': '',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if error is not None:
            _body_params = error


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/errors/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def upload_logs(
        self,
        log_entry: List[LogEntry],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> object:
        """Upload logs

        Upload logs

        :param log_entry: (required)
        :type log_entry: List[LogEntry]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_logs_serialize(
            log_entry=log_entry,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def upload_logs_with_http_info(
        self,
        log_entry: List[LogEntry],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[object]:
        """Upload logs

        Upload logs

        :param log_entry: (required)
        :type log_entry: List[LogEntry]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_logs_serialize(
            log_entry=log_entry,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def upload_logs_without_preload_content(
        self,
        log_entry: List[LogEntry],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Upload logs

        Upload logs

        :param log_entry: (required)
        :type log_entry: List[LogEntry]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_logs_serialize(
            log_entry=log_entry,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _upload_logs_serialize(
        self,
        log_entry,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'LogEntry': '',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if log_entry is not None:
            _body_params = log_entry


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/logs/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def upload_metrics(
        self,
        metric: List[Metric],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> object:
        """Upload metrics

        Upload metrics

        :param metric: (required)
        :type metric: List[Metric]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_metrics_serialize(
            metric=metric,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def upload_metrics_with_http_info(
        self,
        metric: List[Metric],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[object]:
        """Upload metrics

        Upload metrics

        :param metric: (required)
        :type metric: List[Metric]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_metrics_serialize(
            metric=metric,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def upload_metrics_without_preload_content(
        self,
        metric: List[Metric],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Upload metrics

        Upload metrics

        :param metric: (required)
        :type metric: List[Metric]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_metrics_serialize(
            metric=metric,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _upload_metrics_serialize(
        self,
        metric,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'Metric': '',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if metric is not None:
            _body_params = metric


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/metrics/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def upload_profiles(
        self,
        profile: List[Profile],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Dict[str, object]:
        """Upload profiles

        Upload profiles

        :param profile: (required)
        :type profile: List[Profile]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_profiles_serialize(
            profile=profile,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Dict[str, object]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def upload_profiles_with_http_info(
        self,
        profile: List[Profile],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Dict[str, object]]:
        """Upload profiles

        Upload profiles

        :param profile: (required)
        :type profile: List[Profile]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_profiles_serialize(
            profile=profile,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Dict[str, object]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def upload_profiles_without_preload_content(
        self,
        profile: List[Profile],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Upload profiles

        Upload profiles

        :param profile: (required)
        :type profile: List[Profile]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_profiles_serialize(
            profile=profile,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Dict[str, object]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _upload_profiles_serialize(
        self,
        profile,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'Profile': '',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if profile is not None:
            _body_params = profile


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/profiles/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    def upload_spans(
        self,
        span: List[Span],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> object:
        """Upload spans

        Upload spans

        :param span: (required)
        :type span: List[Span]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_spans_serialize(
            span=span,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    def upload_spans_with_http_info(
        self,
        span: List[Span],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[object]:
        """Upload spans

        Upload spans

        :param span: (required)
        :type span: List[Span]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_spans_serialize(
            span=span,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    def upload_spans_without_preload_content(
        self,
        span: List[Span],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Upload spans

        Upload spans

        :param span: (required)
        :type span: List[Span]
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._upload_spans_serialize(
            span=span,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _upload_spans_serialize(
        self,
        span,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'Span': '',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if span is not None:
            _body_params = span


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'APIKeyHeader'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/api/v1/spans/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





import logging
import os
import sys
import cProfile, pstats
import threading
import json

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder

logger = logging.getLogger('graphsignal')

class PythonRecorder(BaseRecorder):
    def __init__(self):
        self._profiler = None
        self._exclude_path = os.path.dirname(os.path.realpath(graphsignal.__file__))

    def _can_include_profile(self, span, profile_name):
        return (graphsignal._tracer.can_include_profile(profile_name) and 
                span.can_include_profile(profile_name))

    def on_span_start(self, span, context):
        if (self._can_include_profile(span, 'profile.cpython') and 
            graphsignal._tracer.set_profiling_mode('profile.cpython', span.name)):

            context['profile.cpython'] = True
            span.set_sampled(True)
            span.set_attribute('sampling.reason', 'profile.cpython')

            if self._profiler:
                # In case of previous profiling not stopped
                self._profiler.disable()
                self._profiler = None
            self._profiler = cProfile.Profile()
            self._profiler.enable()

    def on_span_stop(self, span, context):
        if context.get('profile.cpython', False):
            graphsignal._tracer.unset_profiling_mode()
            if self._profiler:
                self._profiler.disable()

    def on_span_read(self, span, context):
        if context.get('profile.cpython', False):
            if self._profiler:
                try:
                    self._convert_to_profile(span)
                finally:
                    self._profiler = None

    def _convert_to_profile(self, span):
        stats = pstats.Stats(self._profiler)

        stats.sort_stats('tottime')
        func_keys = stats.fcn_list[:] if stats.fcn_list else list(stats.stats.keys())

        func_list = []
        visited = {}
        for func_key in func_keys:
            func_stats = stats.stats[func_key]
            file_name, line_num, func_name = func_key
            if self._has_exclude_func(stats.stats, func_key, visited):
                continue
            cc, nc, tt, ct, _ = func_stats
            func_list.append((file_name, line_num, func_name, nc, tt, ct))

        wt_profile = []
        for file_name, line_num, func_name, nc, tt, ct in func_list[:250]:
            wt_profile.append(dict(
                func_name = _format_frame(file_name, line_num, func_name),
                count = int(nc),
                wall_time_ns = _ns(ct),
                self_wall_time_ns = _ns(tt),
            ))

        span.set_profile(
            name='profile.cpython',
            format='event-averages', 
            content=json.dumps(wt_profile))


    def _has_exclude_func(self, stats, func_key, visited):
        if func_key in visited:
            return visited[func_key]

        file_name, _, _ = func_key
        _, _, _, _, callers = stats[func_key]

        if file_name.startswith(self._exclude_path):
            visited[func_key] = True
            return True
        else:
            visited[func_key] = False

        for caller_key in callers.keys():
            if self._has_exclude_func(stats, caller_key, visited):
                visited[func_key] = True
                return True

        return False


def _ns(sec):
    return int(sec * 1e9)


def _format_frame(file_name, line_num, func_name):
    if file_name == '~':
        file_name = ''

    if file_name and line_num and func_name:
        return '{func_name} ({file_name}:{line_num})'.format(
            file_name=file_name,
            func_name=func_name,
            line_num=line_num)
    elif file_name and func_name:
        return '{func_name} ({file_name})'.format(
            file_name=file_name,
            func_name=func_name)
    elif func_name:
        return func_name
    else:
        return 'unknown'


import os
import logging
import time

try:
    from prometheus_client import REGISTRY
    PROMETHEUS_AVAILABLE = True
except ImportError:
    PROMETHEUS_AVAILABLE = False
    REGISTRY = None

import graphsignal

logger = logging.getLogger('graphsignal')


class PrometheusAdapter():
    def __init__(self, registry=None, name_map_func=None):
        if not PROMETHEUS_AVAILABLE:
            logger.warning("Prometheus client not available. Prometheus metrics will not be collected.")
            return

        if registry is None:
            self._registry = REGISTRY
        else:
            self._registry = registry
        self._last_values = {}

        if not name_map_func:
            raise ValueError('name_map_func is required')
        self._name_map_func = name_map_func

    def setup(self):
        pass

    def collect(self):
        if not PROMETHEUS_AVAILABLE:
            return

        store = graphsignal._tracer.metric_store()
        metric_tags = graphsignal._tracer.tags.copy()
        now = int(time.time())

        try:
            metrics_collected = 0

            for metric in self._registry.collect():
                target_name = self._name_map_func(metric.name)
                if not target_name:
                    continue
                
                metrics_collected += 1
                
                sample_tags = metric_tags.copy()
                
                # Get labels from the first sample (they are consistent across sample group)
                if metric.samples:
                    for label_name, label_value in metric.samples[0].labels.items():
                        if label_name not in ['le', 'quantile']:
                            sample_tags[label_name] = label_value

                sample_groups = {}
                for sample in metric.samples:
                    group_key = frozenset(sample.labels.items()) if sample.labels else frozenset()
                    if group_key not in sample_groups:
                        sample_groups[group_key] = {}
                    sample_groups[group_key][sample.name] = sample
                
                for sample_map in sample_groups.values():
                    if metric.type == 'gauge':
                        sample = sample_map.get(metric.name, None)
                        if sample:
                            store.set_gauge(
                                name=target_name, 
                                tags=sample_tags, 
                                value=sample.value, 
                                update_ts=now)
                    
                    elif metric.type == 'counter':
                        sample = sample_map.get(f'{metric.name}_total', None)
                        if sample:
                            metric_key = (target_name, tuple(sorted(sample_tags.items())))
                            current_value = sample.value
                            
                            if metric_key in self._last_values:
                                last_value = self._last_values[metric_key]
                                delta = current_value - last_value
                                if delta >= 0:
                                    store.inc_counter(
                                        name=target_name, 
                                        tags=sample_tags, 
                                        value=delta, 
                                        update_ts=now)
                            else:
                                # First time seeing this counter, store initial value
                                self._last_values[metric_key] = current_value
                            
                            # Update the last value for next delta calculation
                            self._last_values[metric_key] = current_value

                    elif metric.type == 'histogram':
                        count_sample = sample_map.get(f'{metric.name}_count', None)
                        sum_sample = sample_map.get(f'{metric.name}_sum', None)
                        if count_sample and sum_sample:
                            store.update_summary(
                                name=target_name, 
                                tags=sample_tags, 
                                count=count_sample.value, 
                                sum_val=sum_sample.value,
                                sum2_val=sum_sample.value * sum_sample.value,  # Approximate sum2
                                update_ts=now)

                    elif metric.type == 'summary':
                        count_sample = sample_map.get(f'{metric.name}_count', None)
                        sum_sample = sample_map.get(f'{metric.name}_sum', None)
                        if count_sample and sum_sample:
                            store.update_summary(
                                name=target_name, 
                                tags=sample_tags, 
                                count=count_sample.value, 
                                sum_val=sum_sample.value,
                                sum2_val=sum_sample.value * sum_sample.value,  # Approximate sum2
                                update_ts=now)

        except Exception as e:
            logger.error(f"Failed to collect Prometheus metrics: {e}", exc_info=True)



# coding: utf-8

"""
    Graphsignal API

    API for uploading and querying spans, errors, metrics, and logs.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from graphsignal.client.models.attribute import Attribute
from graphsignal.client.models.counter import Counter
from graphsignal.client.models.tag import Tag
from typing import Optional, Set
from typing_extensions import Self

class Span(BaseModel):
    """
    Span
    """ # noqa: E501
    span_id: StrictStr = Field(description="Unique identifier for the span.")
    trace_id: StrictStr = Field(description="Identifier of the trace.")
    parent_span_id: Optional[StrictStr] = Field(default=None, description="Identifier of the parent span, if this is a nested span.")
    linked_span_ids: Optional[List[StrictStr]] = Field(default=None, description="List of linked span identifiers.")
    start_ns: StrictInt = Field(description="Start time of the span in nanoseconds.")
    end_ns: StrictInt = Field(description="End time of the span in nanoseconds.")
    name: StrictStr = Field(description="The name of the span.")
    tags: Optional[List[Tag]] = Field(default=None, description="List of tags associated with the span.")
    attributes: Optional[List[Attribute]] = Field(default=None, description="List of attributes associated with the span.")
    counters: Optional[List[Counter]] = Field(default=None, description="Counters associated with the span.")
    __properties: ClassVar[List[str]] = ["span_id", "trace_id", "parent_span_id", "linked_span_ids", "start_ns", "end_ns", "name", "tags", "attributes", "counters"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Span from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in tags (list)
        _items = []
        if self.tags:
            for _item_tags in self.tags:
                if _item_tags:
                    _items.append(_item_tags.to_dict())
            _dict['tags'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in attributes (list)
        _items = []
        if self.attributes:
            for _item_attributes in self.attributes:
                if _item_attributes:
                    _items.append(_item_attributes.to_dict())
            _dict['attributes'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in counters (list)
        _items = []
        if self.counters:
            for _item_counters in self.counters:
                if _item_counters:
                    _items.append(_item_counters.to_dict())
            _dict['counters'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Span from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "span_id": obj.get("span_id"),
            "trace_id": obj.get("trace_id"),
            "parent_span_id": obj.get("parent_span_id"),
            "linked_span_ids": obj.get("linked_span_ids"),
            "start_ns": obj.get("start_ns"),
            "end_ns": obj.get("end_ns"),
            "name": obj.get("name"),
            "tags": [Tag.from_dict(_item) for _item in obj["tags"]] if obj.get("tags") is not None else None,
            "attributes": [Attribute.from_dict(_item) for _item in obj["attributes"]] if obj.get("attributes") is not None else None,
            "counters": [Counter.from_dict(_item) for _item in obj["counters"]] if obj.get("counters") is not None else None
        })
        return _obj





# coding: utf-8

"""
    Graphsignal API

    API for uploading and querying spans, errors, metrics, and logs.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class MetricType(str, Enum):
    """
    MetricType
    """

    """
    allowed enum values
    """
    GAUGE = 'gauge'
    COUNTER = 'counter'
    SUMMARY = 'summary'
    HISTOGRAM = 'histogram'

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of MetricType from a JSON string"""
        return cls(json.loads(json_str))





import logging
import os
import json
import time
import torch

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.profiler_utils import create_log_dir, remove_log_dir

logger = logging.getLogger('graphsignal')

class PyTorchRecorder(BaseRecorder):
    def __init__(self):
        self._torch_prof = None
        self._log_dir = None

    def setup(self):
        tracer = graphsignal._tracer
        tracer.set_tag('framework.name', 'pytorch')
        tracer.set_tag('framework.version', torch.__version__)

    def _can_include_profile(self, span, profile_name):
        return (graphsignal._tracer.can_include_profile(profile_name) and 
                span.can_include_profile(profile_name))

    def on_span_start(self, span, context):
        if (self._can_include_profile(span, 'profile.pytorch') and 
            graphsignal._tracer.set_profiling_mode('profile.pytorch', span.name)):
            context['profile.pytorch'] = True
            span.set_sampled(True)
            span.set_attribute('sampling.reason', 'profile.pytorch')

            if self._torch_prof:
                # In case of previous profiling not stopped
                self._torch_prof.stop()
                self._torch_prof = None

            def _schedule_func(step):
                return torch.profiler.ProfilerAction.RECORD
            self._torch_prof = torch.profiler.profile(
                activities=[
                    torch.profiler.ProfilerActivity.CPU, 
                    torch.profiler.ProfilerActivity.CUDA],
                schedule=_schedule_func,
                record_shapes=False,
                profile_memory=True,
                with_stack=False,
                with_flops=True)
            self._torch_prof.start()

    def on_span_stop(self, span, context):
        if context.get('profile.pytorch', False):
            graphsignal._tracer.unset_profiling_mode()
            if self._torch_prof:
                self._torch_prof.stop()

    def on_span_read(self, span, context):
        if context.get('profile.pytorch', False):
            if self._torch_prof:
                try:
                    self._convert_to_profile(span)
                except Exception as e:
                    logger.debug('Error converting to profile', exc_info=True)
                finally:
                    self._torch_prof = None

    def on_metric_update(self):
        if not torch.cuda.is_available():
            return

        now = int(time.time())
        device_count = torch.cuda.device_count()
        
        for device_idx in range(device_count):
            try:
                # Get device properties
                device_props = torch.cuda.get_device_properties(device_idx)
                device_name = device_props.name
                
                # Get memory stats for this device
                memory_stats = torch.cuda.memory_stats(device_idx)
                
                # Set up metric tags
                store = graphsignal._tracer.metric_store()
                metric_tags = graphsignal._tracer.tags.copy()
                metric_tags['device.type'] = 'gpu'
                metric_tags['device.index'] = device_idx
                metric_tags['device.name'] = device_name
                metric_tags['framework.name'] = 'pytorch'
                metric_tags['framework.version'] = torch.__version__
                
                # Record memory metrics
                if 'allocated_bytes.all.current' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.allocated', tags=metric_tags,
                        value=memory_stats['allocated_bytes.all.current'], update_ts=now)
                
                if 'reserved_bytes.all.current' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.reserved', tags=metric_tags,
                        value=memory_stats['reserved_bytes.all.current'], update_ts=now)
                
                if 'allocated_bytes.all.peak' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.allocated.peak', tags=metric_tags,
                        value=memory_stats['allocated_bytes.all.peak'], update_ts=now)
                
                if 'reserved_bytes.all.peak' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.reserved.peak', tags=metric_tags,
                        value=memory_stats['reserved_bytes.all.peak'], update_ts=now)
                
                # Record allocation/deallocation counts
                if 'allocated_bytes.all.count' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.allocations', tags=metric_tags,
                        value=memory_stats['allocated_bytes.all.count'], update_ts=now)
                
                if 'allocated_bytes.all.freed' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.deallocations', tags=metric_tags,
                        value=memory_stats['allocated_bytes.all.freed'], update_ts=now)
                
                # Record additional memory management metrics
                if 'num_alloc_retries' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.alloc_retries', tags=metric_tags,
                        value=memory_stats['num_alloc_retries'], update_ts=now)
                
                if 'num_ooms' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.ooms', tags=metric_tags,
                        value=memory_stats['num_ooms'], update_ts=now)
                
                if 'num_sync_all_streams' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.sync_all_streams', tags=metric_tags,
                        value=memory_stats['num_sync_all_streams'], update_ts=now)
                
                if 'num_device_alloc' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.device_alloc', tags=metric_tags,
                        value=memory_stats['num_device_alloc'], update_ts=now)
                
                if 'num_device_free' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.device_free', tags=metric_tags,
                        value=memory_stats['num_device_free'], update_ts=now)
                
                # Record fragmentation metrics
                if 'allocated_bytes.all.peak' in memory_stats and 'reserved_bytes.all.peak' in memory_stats:
                    if memory_stats['reserved_bytes.all.peak'] > 0:
                        fragmentation = (memory_stats['reserved_bytes.all.peak'] - memory_stats['allocated_bytes.all.peak']) / memory_stats['reserved_bytes.all.peak']
                        store.set_gauge(
                            name='pytorch.memory.fragmentation', tags=metric_tags,
                            value=fragmentation * 100, update_ts=now, unit='percent')
                
                # Record cache metrics
                if 'allocated_bytes.all.cached' in memory_stats:
                    store.set_gauge(
                        name='pytorch.memory.cached', tags=metric_tags,
                        value=memory_stats['allocated_bytes.all.cached'], update_ts=now)
                
                # Record device memory info
                device_memory = torch.cuda.get_device_properties(device_idx).total_memory
                if device_memory > 0:
                    store.set_gauge(
                        name='pytorch.memory.total', tags=metric_tags,
                        value=device_memory, update_ts=now)
                
                # Calculate memory utilization percentage
                if device_memory > 0 and 'reserved_bytes.all.current' in memory_stats:
                    utilization = (memory_stats['reserved_bytes.all.current'] / device_memory) * 100
                    store.set_gauge(
                        name='pytorch.memory.utilization', tags=metric_tags,
                        value=utilization, update_ts=now, unit='percent')
                
            except Exception as e:
                logger.warning(f'Failed to record PyTorch memory metrics for device {device_idx}: {e}')

    def _convert_to_profile(self, span):
        cpu_profile = []
        for event_avg in self._torch_prof.key_averages():
            if event_avg.key and event_avg.key.startswith('ProfilerStep'):
                continue
            cpu_profile.append(dict(
                op_name = event_avg.key,
                device_type = event_avg.device_type.name if event_avg.device_type else None,
                count = _uint(event_avg.count),
                cpu_time_ns = _ns(event_avg.cpu_time_total),
                self_cpu_time_ns = _ns(event_avg.self_cpu_time_total),
                device_time_ns = _ns(event_avg.device_time_total),
                self_device_time_ns = _ns(event_avg.self_device_time_total),
                cpu_memory = _uint(event_avg.cpu_memory_usage),
                self_cpu_memory = _uint(event_avg.self_cpu_memory_usage),
                device_memory = _uint(event_avg.device_memory_usage),
                self_device_memory = _uint(event_avg.self_device_memory_usage),
                flops = _uint(event_avg.flops)
            ))

        if len(cpu_profile) > 0:
            span.set_profile(
                name='profile.pytorch.cpu', 
                format='event-averages', 
                content=json.dumps(cpu_profile))

        kernel_index = {}
        for event in self._torch_prof.events():
            for kernel in event.kernels:
                key = (event.key, kernel.name, kernel.device)
                if key in kernel_index:
                    kernel_avg = kernel_index[key]
                    kernel_avg['count'] += 1
                    kernel_avg['duration_ns'] += _ns(kernel.duration)
                else:
                    kernel_avg = kernel_index[key] = dict(
                        device_idx = kernel.device,
                        op_name = event.name,
                        kernel_name = kernel.name,
                        count = 1,
                        duration_ns = _ns(kernel.duration)
                    )

        device_profile = list(kernel_index.values())
        span.set_profile(
            name='profile.pytorch.kernel', 
            format='event-averages', 
            content=json.dumps(device_profile))

        chrome_trace = self._export_chrome_trace()
        if chrome_trace:
            span.set_profile('profile.pytorch.trace', 'chrome-trace', chrome_trace)

    def _export_chrome_trace(self):
        try:
            read_start_time = time.time()

            self._log_dir = create_log_dir()

            trace_path = os.path.join(self._log_dir, 'trace.json')
            self._torch_prof.export_chrome_trace(trace_path)

            trace_file_size = os.path.getsize(trace_path)
            logger.debug('Chrome trace size: %s', trace_file_size)
            if trace_file_size > 500 * 1e6:
                logger.debug('Chrome trace file too big: %s', trace_file_size)
                return None

            with open(trace_path, "r") as f:
                return str(f.read())
        finally:
            remove_log_dir(self._log_dir)
            logger.debug('Chrome trace export time: %s', time.time() - read_start_time)

        return None

def _ns(val):
    return int(max(val, 0) * 1e3)


def _uint(val):
    return max(int(val), 0)



import logging
import sys
import threading
import time
import math

import graphsignal
from graphsignal import version
from graphsignal import client

logger = logging.getLogger('graphsignal')


class BaseMetric:
    def __init__(self, name, tags, unit=None):
        self._update_lock = threading.Lock()
        self.name = name
        self.tags = tags
        self.unit = unit
        self.is_updated = False

    def touch(self):
        self.is_updated = True

    def export(self):
        self.is_updated = False

        model = client.Metric(
            name=self.name,
            tags=[],
            type=self.type,
            update_ts=self.update_ts)
        for key, value in self.tags.items():
            model.tags.append(client.Tag(
                key=str(key)[:50],
                value=str(value)[:250]
            ))
        if self.unit is not None:
            model.unit = self.unit
        return model


class GaugeMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'gauge'
        self.gauge = None

    def update(self, value, update_ts):
        with self._update_lock:
            self.touch()
            self.gauge = value
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            model = super().export()
            model.gauge = self.gauge
            self.gauge = None
            return model


class CounterMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'counter'
        self.total = 0

    def update(self, value, update_ts):
        with self._update_lock:
            self.touch()
            self.total += value
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            model = super().export()
            model.total = self.total
            self.total = 0
            return model


class SummaryMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'summary'
        self.count = 0
        self.sum = 0
        self.sum2 = 0

    def update(self, count, sum_val, sum2_val, update_ts):
        with self._update_lock:
            self.touch()
            self.count += count
            self.sum += sum_val
            self.sum2 += sum2_val
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            model = super().export()
            model.summary = client.Summary(count=int(self.count), sum=self.sum, sum2=self.sum2)
            return model


class HistogramMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'histogram'
        self.histogram = {}

    def update(self, value, update_ts):
        if value < 1:
            return
        with self._update_lock:
            self.touch()
            bin_size = max(10 ** (int(math.log(value, 10)) - 1), 1)
            bin = int(value / bin_size) * bin_size
            self.histogram[bin] = self.histogram.get(bin, 0) + 1
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            model = super().export()
            model.histogram = client.Histogram(bins=[], counts=[])
            for bin, count in self.histogram.items():
                model.histogram.bins.append(bin)
                model.histogram.counts.append(int(count))
            self.histogram = {}
            return model


class MetricStore:
    def __init__(self):
        self._update_lock = threading.Lock()
        self._has_unexported = False
        self._metrics = {}

    def metric_key(self, name, tags):
        return (name, frozenset(tags.items()))

    def set_gauge(self, name, tags, value, update_ts, unit=None):
        if name is None:
            raise ValueError('Gauge name cannot be None')
        if value is None:
            raise ValueError('Gauge value cannot be None')

        key = self.metric_key(name, tags)
        with self._update_lock:
            if key not in self._metrics:
                metric = self._metrics[key] = GaugeMetric(name, tags, unit=unit)
            else:
                metric = self._metrics[key]
        metric.update(value, update_ts)
        return metric

    def inc_counter(self, name, tags, value, update_ts, unit=None):
        if name is None:
            raise ValueError('Counter name cannot be None')
        if value is None:
            raise ValueError('Counter value cannot be None')
        
        key = self.metric_key(name, tags)
        with self._update_lock:
            if key not in self._metrics:
                metric = self._metrics[key] = CounterMetric(name, tags, unit=unit)
            else:
                metric = self._metrics[key]
        metric.update(value, update_ts)
        return metric

    def update_summary(self, name, tags, count, sum_val, sum2_val, update_ts, unit=None):
        if name is None:
            raise ValueError('Summary name cannot be None')
        if count is None:
            raise ValueError('Summary count cannot be None')
        if sum_val is None:
            raise ValueError('Summary sum cannot be None')
        if sum2_val is None:
            raise ValueError('Summary sum2 cannot be None')
        
        key = self.metric_key(name, tags)
        with self._update_lock:
            if key not in self._metrics:
                metric = self._metrics[key] = SummaryMetric(name, tags, unit=unit)
            else:
                metric = self._metrics[key]
        metric.update(count, sum_val, sum2_val, update_ts)
        return metric

    def update_histogram(self, name, tags, value, update_ts, unit=None):
        if name is None:
            raise ValueError('Histogram name cannot be None')
        if value is None:
            raise ValueError('Histogram value cannot be None')

        key = self.metric_key(name, tags)
        with self._update_lock:
            if key not in self._metrics:
                metric = self._metrics[key] = HistogramMetric(name, tags, unit=unit)
            else:
                metric = self._metrics[key]
        metric.update(value, update_ts)
        return metric

    def has_unexported(self):
        with self._update_lock:
            for metric in self._metrics.values():
                if metric.is_updated:
                    return True
        return False

    def export(self):
        models = []
        with self._update_lock:
            for metric in self._metrics.values():
                if metric.is_updated:
                    models.append(metric.export())
        return models

    def clear(self):
        with self._update_lock:
            self._metrics = {}


import logging
from functools import wraps
import asyncio
import types
import re
import time

import graphsignal

logger = logging.getLogger('graphsignal')

version_regexp = re.compile(r'^(\d+)\.?(\d+)?\.?(\d+)?')


def profile_method(obj, func_name, event_name=None, event_name_func=None, profile_func=None):
    def before_func(args, kwargs):
        profiled_event_name = None
        if event_name:
            profiled_event_name = event_name
        elif event_name_func:
            try:
                profiled_event_name = event_name_func(args, kwargs)
            except Exception as e:
                logger.debug('Error tracing %s', func_name, exc_info=True)
        else:
            profiled_event_name = func_name

        return dict(
            event_name=profiled_event_name,
            start_ns=time.perf_counter_ns())

    def after_func(args, kwargs, ret, exc, context):
        start_ns = context['start_ns']

        if not is_generator(ret) and not is_async_generator(ret):
            duration_ns = time.perf_counter_ns() - start_ns

            try:
                if not context.get('finished', False):
                    context['finished'] = True
                    profile_func(context['event_name'], duration_ns)
            except Exception as e:
                logger.debug('Error tracing %s', func_name, exc_info=True)

    def yield_func(stopped, item, context, exc):
        start_ns = context['start_ns']

        if stopped:
            duration_ns = time.perf_counter_ns() - start_ns

            try:
                if not context.get('finished', False):
                    context['finished'] = True
                    profile_func(context['event_name'], duration_ns)
            except Exception as e:
                logger.debug('Error tracing %s', func_name, exc_info=True)

    if not patch_method(obj, func_name, before_func=before_func, after_func=after_func, yield_func=yield_func):
        logger.debug('Cannot instrument %s.', func_name)


def trace_method(obj, func_name, span_name=None, include_profiles=None, span_name_func=None, trace_func=None, data_func=None):
    def before_func(args, kwargs):
        if span_name_func is None:
            selected_span_name = span_name
        else:
            selected_span_name = span_name_func(args, kwargs)
        return dict(span=graphsignal.trace(span_name=selected_span_name, include_profiles=include_profiles))

    def after_func(args, kwargs, ret, exc, context):
        span = context['span']

        if not is_generator(ret) and not is_async_generator(ret):
            span.measure()

        try:
            if exc is not None:
                span.add_exception(exc)

            trace_func(span, args, kwargs, ret, exc)
        except Exception as e:
            logger.debug('Error tracing %s', func_name, exc_info=True)

        if not is_generator(ret) and not is_async_generator(ret):
            span.stop()

    def yield_func(stopped, item, context, exc):
        span = context['span']

        if exc is not None:
            span.add_exception(exc)

        if stopped:
            span.stop()
            if data_func:
                data_func(span, None, exc, stopped)
        else:
            if data_func:
                data_func(span, item, exc, stopped)

    if not patch_method(obj, func_name, before_func=before_func, after_func=after_func, yield_func=yield_func):
        logger.debug('Cannot instrument %s.', func_name)


def uninstrument_method(obj, func_name):
    if not unpatch_method(obj, func_name):
        logger.debug('Cannot uninstrument %s.', func_name)


class ObjectProxy:
    def __init__(self, obj):
        self._obj = obj

    def __getattr__(self, attr):
        return getattr(self._obj, attr)

    def __repr__(self):
        return repr(self._obj)


class GeneratorWrapper(ObjectProxy):
    def __init__(self, gen, yield_func, context=None):
        super().__init__(gen)
        self._gen = gen
        self._yield_func = yield_func
        self._context = context

    def __enter__(self):
        if hasattr(self._gen, '__enter__'):
            self._gen.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._gen, '__exit__'):
            self._gen.__exit__(exc_type, exc_val, exc_tb)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = self._gen.__next__()
            try:
                self._yield_func(False, item, self._context, None)
            except:
                logger.debug('Exception in yield_func', exc_info=True)
            return item
        except StopIteration:
            try:
                self._yield_func(True, None, self._context, None)
            except:
                logger.debug('Exception in yield_func', exc_info=True)
            raise
        except Exception as ex:
            try:
                self._yield_func(True, None, self._context, ex)
            except:
                logger.debug('Exception in yield_func', exc_info=True)
            raise


class AsyncGeneratorWrapper(ObjectProxy):
    def __init__(self, async_gen, yield_func, context=None):
        super().__init__(async_gen)
        self._async_gen = async_gen
        self._yield_func = yield_func
        self._context = context

    async def __aenter__(self):
        if hasattr(self._async_gen, '__aenter__'):
            await self._async_gen.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._async_gen, '__aexit__'):
            await self._async_gen.__aexit__(exc_type, exc_val, exc_tb)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            item = await self._async_gen.__anext__()
            try:
                self._yield_func(False, item, self._context, None)
            except:
                logger.debug('Exception in yield_func', exc_info=True)
            return item
        except StopAsyncIteration:
            try:
                self._yield_func(True, None, self._context, None)
            except:
                logger.debug('Exception in yield_func', exc_info=True)
            raise
        except Exception as ex:
            try:
                self._yield_func(True, None, self._context, ex)
            except:
                logger.debug('Exception in yield_func', exc_info=True)
            raise


def patch_method(obj, func_name, before_func=None, after_func=None, yield_func=None):
    if not hasattr(obj, func_name):
        return False

    func = getattr(obj, func_name)

    if hasattr(func, '__graphsignal_wrapped__'):
        return False

    if asyncio.iscoroutinefunction(func) or asyncio.iscoroutinefunction(getattr(func, '__wrapped__', None)):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            context = None
            exc = None
            ret = None

            if before_func:
                try:
                    context = before_func(args, kwargs)
                except:
                    logger.debug('Exception in before_func', exc_info=True)

            try:
                ret = await func(*args, **kwargs)
            except BaseException as e:
                exc = e

            if after_func:
                try:
                    after_func(args, kwargs, ret, exc, context)
                except:
                    logger.debug('Exception in after_func', exc_info=True)

            if yield_func:
                try:
                    if is_generator(ret):
                        ret = GeneratorWrapper(ret, yield_func, context)
                    elif is_async_generator(ret):
                        ret = AsyncGeneratorWrapper(ret, yield_func, context)
                except:
                    logger.debug('Exception in yield_func', exc_info=True)

            if exc:
                raise exc
            return ret
    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            context = None
            exc = None
            ret = None

            if before_func:
                try:
                    context = before_func(args, kwargs)
                except:
                    logger.debug('Exception in before_func', exc_info=True)

            try:
                ret = func(*args, **kwargs)
            except BaseException as e:
                exc = e

            if after_func:
                try:
                    after_func(args, kwargs, ret, exc, context)
                except:
                    logger.debug('Exception in after_func', exc_info=True)

            if yield_func:
                try:
                    if is_generator(ret):
                        ret = GeneratorWrapper(ret, yield_func, context)
                    elif is_async_generator(ret):
                        ret = AsyncGeneratorWrapper(ret, yield_func, context)
                except:
                    logger.debug('Exception in yield_func', exc_info=True)

            if exc:
                raise exc
            return ret

    setattr(wrapper, '__graphsignal_wrapped__', True)
    setattr(obj, func_name, wrapper)
    return True


def unpatch_method(obj, func_name):
    if not hasattr(obj, func_name):
        return False

    func = getattr(obj, func_name)

    if not hasattr(func, '__graphsignal_wrapped__'):
        return False

    if not hasattr(func, '__wrapped__'):
        return False
    
    setattr(obj, func_name, getattr(func, '__wrapped__'))
    return True


def is_generator(obj):
    if not obj:
        return False
    if isinstance(obj, types.GeneratorType):
        return True
    if hasattr(obj, '__iter__') and hasattr(obj, '__next__'):
        return True
    return False


def is_async_generator(obj):
    if not obj:
        return False
    if isinstance(obj, types.AsyncGeneratorType):
        return True
    if hasattr(obj, '__aiter__') and hasattr(obj, '__anext__'):
        return True
    return False


def read_args(args, kwargs, names):
    values = {}
    for name, arg in zip(names, args):
        values[name] = arg
    values.update(kwargs)
    return values


def parse_semver(version):
    parsed_version = [0, 0, 0]
    version_match = version_regexp.match(str(version))
    if version_match is not None:
        groups = version_match.groups()
        if groups[0] is not None:
            parsed_version[0] = int(groups[0])
        if groups[1] is not None:
            parsed_version[1] = int(groups[1])
        if groups[2] is not None:
            parsed_version[2] = int(groups[2])
    return tuple(parsed_version)


def compare_semver(version1, version2):
    version1_int = version1[0] * 1e6 + version1[1] * 1e3 + version1[2]
    version2_int = version2[0] * 1e6 + version2[1] * 1e3 + version2[2]
    if version1_int < version2_int:
        return -1
    if version1_int > version2_int:
        return 1
    else:
        return 0


from token import OP
from typing import Dict, Optional
import logging
import sys
import os
import time
import traceback
import importlib
import socket
import contextvars
import importlib.abc
import threading
import functools
import asyncio
from pathlib import Path

from graphsignal.uploader import Uploader
from graphsignal.metrics import MetricStore
from graphsignal.logs import LogStore
from graphsignal.spans import Span
from graphsignal import client
from graphsignal.utils import uuid_sha1, sanitize_str

logger = logging.getLogger('graphsignal')



class GraphsignalTracerLogHandler(logging.Handler):
    def __init__(self, tracer):
        super().__init__()
        self._tracer = tracer

    def emit(self, record):
        try:
            log_tags = self._tracer.tags.copy()

            exception = None
            if record.exc_info and isinstance(record.exc_info, tuple):
                exception = self.format(record)

            self._tracer.log_store().log_tracer_message(
                tags=log_tags,
                level=record.levelname,
                message=record.getMessage(),
                exception=exception)
        except Exception:
            pass


RECORDER_SPECS = {
    '(default)': [
        ('graphsignal.recorders.process_recorder', 'ProcessRecorder'),
        ('graphsignal.recorders.python_recorder', 'PythonRecorder'),
        ('graphsignal.recorders.exception_recorder', 'ExceptionRecorder'),
        ('graphsignal.recorders.nvml_recorder', 'NVMLRecorder')],
    'torch': [('graphsignal.recorders.pytorch_recorder', 'PyTorchRecorder')],
    'vllm': [('graphsignal.recorders.vllm_recorder', 'VLLMRecorder')]
}

class SourceLoaderWrapper(importlib.abc.SourceLoader):
    def __init__(self, loader, tracer):
        self._loader = loader
        self._tracer = tracer

    def create_module(self, spec):
        return self._loader.create_module(spec)

    def exec_module(self, module):
        self._loader.exec_module(module)
        try:
            self._tracer.initialize_recorders_for_module(module.__name__)
        except Exception:
            logger.error('Error initializing recorders for module %s', module.__name__, exc_info=True)

    def load_module(self, fullname):
        self._loader.load_module(fullname)

    def get_data(self, path):
        return self._loader.get_data(path)

    def get_filename(self, fullname):
        return self._loader.get_filename(fullname)


class SupportedModuleFinder(importlib.abc.MetaPathFinder):
    def __init__(self, tracer):
        self._tracer = tracer
        self._disabled = False

    def find_spec(self, fullname, path=None, target=None):
        if self._disabled:
            return None

        if fullname in RECORDER_SPECS:
            try:
                self._disabled = True
                spec = importlib.util.find_spec(fullname)
                if spec:
                    loader = importlib.util.find_spec(fullname).loader
                    if loader is not None and isinstance(loader, importlib.abc.SourceLoader):
                        return importlib.util.spec_from_loader(fullname, SourceLoaderWrapper(loader, self._tracer))
            except Exception:
                logger.error('Error patching spec for module %s', fullname, exc_info=True)
            finally:
                self._disabled = False

        return None

class SamplingTokenBucket:
    def __init__(self, sampling_rate_per_minute: float):
        self.capacity = sampling_rate_per_minute   # max tokens (samples) per minute
        self.tokens = self.capacity               # start full
        self.refill_rate_per_sec = self.capacity / 60.0  # tokens per second
        self.last_refill_time = time.monotonic()
    
    def _refill(self):
        now = time.monotonic()
        elapsed = now - self.last_refill_time
        if elapsed > 0:
            # Add tokens based on elapsed time
            new_tokens = elapsed * self.refill_rate_per_sec
            self.tokens = min(self.capacity, self.tokens + new_tokens)
            self.last_refill_time = now
    
    def should_sample(self, consume=True) -> bool:
        # Only refill if we don't have enough tokens
        if self.tokens < 1:
            self._refill()
        
        if self.tokens >= 1:
            if consume:
                self.tokens -= 1
            return True
        return False

class DynamicCycle:
    def __init__(self):
        self.keys = []
        self.index = 0

    def add(self, key):
        if key not in self.keys:
            self.keys.append(key)

    def next(self):
        if not self.keys:
            raise ValueError("No keys to sample from")
        key = self.keys[self.index]
        self.index = (self.index + 1) % len(self.keys)
        return key

class Tracer:
    TICK_INTERVAL_SEC = 10
    MAX_PROCESS_TAGS = 25
    PROFILING_MODE_TIMEOUT_SEC = 60
    MAX_ERRORS_PER_MINUTE = 25
    VALID_ERROR_LEVELS = {'debug', 'info', 'warning', 'error', 'critical'}
    MAX_SAMPLERS = 100

    def __init__(
            self, 
            api_key=None, 
            api_url=None, 
            tags=None, 
            auto_instrument=True, 
            samples_per_min=5,
            profiles_per_min=1,
            include_profiles=None,
            debug_mode=False):
        if debug_mode:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.WARNING)
        
        if not api_key:
            raise ValueError('api_key is required')

        self.api_key = api_key
        if api_url:
            self.api_url = api_url
        else:
            self.api_url = 'https://api.graphsignal.com'
        self.tags = {}
        if tags:
            self.tags.update(tags)
        self.context_tags = contextvars.ContextVar('graphsignal_context_tags', default={})

        self.auto_instrument = auto_instrument
        self.samples_per_min = samples_per_min if samples_per_min is not None else 5
        self.profiles_per_min = profiles_per_min if profiles_per_min is not None else 1
        self.include_profiles = include_profiles
        self.debug_mode = debug_mode

        self._sampling_token_buckets = {}

        self._error_counter = 0
        self._error_counter_reset_time = time.time()

        self._tick_timer_thread = None
        self._tick_stop_event = threading.Event()
        self._tick_lock = threading.Lock()
        self._tick_run_thread = None
        self._tracer_log_handler = None
        self._uploader = None
        self._metric_store = None
        self._log_store = None
        self._recorders = None
        self._profiling_mode_lock = threading.Lock()
        self._profiling_mode = None
        self._include_profiles_index = set(include_profiles) if isinstance(include_profiles, list) else None
        self._active_profiles = DynamicCycle()

        self._process_start_ms = int(time.time() * 1e3)

        self.last_tick_ts = time.time()

        self.auto_export = True

    def setup(self):
        self._uploader = Uploader()
        self._uploader.setup()
        self._metric_store = MetricStore()
        self._log_store = LogStore()
        self._recorders = {}

        # initialize tracer log handler
        self._tracer_log_handler = GraphsignalTracerLogHandler(self)
        logger.addHandler(self._tracer_log_handler)

        # initialize module wrappers
        self._module_finder = SupportedModuleFinder(self)
        sys.meta_path.insert(0, self._module_finder)

        # initialize default recorders and recorders for already loaded modules
        for module_name in RECORDER_SPECS.keys():
            if module_name == '(default)' or module_name in sys.modules:
                self.initialize_recorders_for_module(module_name)

        # start the tick timer thread
        self._start_tick_timer()

    def _start_tick_timer(self):
        def _tick_loop():
            while not self._tick_stop_event.wait(Tracer.TICK_INTERVAL_SEC):
                try:
                    if self.auto_export:
                        self.tick()
                except Exception as exc:
                    logger.error('Error in tick timer: %s', exc, exc_info=True)

        self._tick_timer_thread = threading.Thread(target=_tick_loop, daemon=True)
        self._tick_timer_thread.start()

    def shutdown(self):
        if self.auto_export:
            self.tick(block=True, force=True)

        if self._tick_stop_event:
            self._tick_stop_event.set()

        if self._tick_timer_thread:
            self._tick_timer_thread.join()
            self._tick_timer_thread = None

        if self._tick_run_thread:
            self._tick_run_thread.join()
            self._tick_run_thread = None

        for recorder in self.recorders():
            recorder.shutdown()

        self._recorders = None
        self._metric_store = None
        self._log_store = None
        self._uploader = None

        self.tags = None

        self.context_tags.set({})
        self.context_tags = None

        # remove module wrappers
        if self._module_finder in sys.meta_path:
            sys.meta_path.remove(self._module_finder)
        self._module_finder = None

        # remove tracer log handler
        logger.removeHandler(self._tracer_log_handler)
        self._tracer_log_handler = None

    def uploader(self):
        return self._uploader

    def initialize_recorders_for_module(self, module_name):
        # check already loaded
        if module_name in self._recorders:
            return

        # check if supported
        if module_name not in RECORDER_SPECS:
            return

        # load recorder
        logger.debug('Initializing recorder for module: %s', module_name)
        specs = RECORDER_SPECS[module_name]
        self._recorders[module_name] = []
        for spec in specs:
            try:
                recorder_module = importlib.import_module(spec[0])
                recorder_class = getattr(recorder_module, spec[1])
                recorder = recorder_class()
                recorder.setup()
                self._recorders[module_name].append(recorder)
            except Exception:
                logger.error('Failed to initialize recorder for module: %s', module_name, exc_info=True)

    def recorders(self):
        for recorder_list in self._recorders.values():
            for recorder in recorder_list:
                yield recorder

    def metric_store(self):
        return self._metric_store

    def log_store(self):
        return self._log_store

    def sampler(self, sampler_key, samples_per_min):
        if sampler_key not in self._sampling_token_buckets:
            if len(self._sampling_token_buckets) >= self.MAX_SAMPLERS:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('Maximum number of samplers reached, skipping sampler: %s', sampler_key)
                return None
            self._sampling_token_buckets[sampler_key] = SamplingTokenBucket(samples_per_min)
        return self._sampling_token_buckets[sampler_key]

    def should_sample(self, sampler_key, consume=True):
        sampler = self.sampler(sampler_key, self.samples_per_min)
        if not sampler:
            return False
        return sampler.should_sample(consume)

    def should_profile(self, sampler_key, consume=True):
        sampler = self.sampler(sampler_key, self.profiles_per_min)
        if not sampler:
            return False
        return sampler.should_sample(consume)

    def set_profiling_mode(self, profile_name, span_name=None):
        # avoid entering the lock for every profiling mode check
        # only check if we can sample, don't consume a token
        sampler_key = (profile_name, span_name) if span_name else profile_name
        if not self.should_profile(sampler_key, consume=False):
            return False

        # allow all profilers to run
        self._active_profiles.add(profile_name)
        if self._active_profiles.next() != profile_name:
            return False
        
        with self._profiling_mode_lock:
            if self._profiling_mode and (time.time() - self._profiling_mode) > self.PROFILING_MODE_TIMEOUT_SEC:
                self._profiling_mode = None

            if self._profiling_mode:
                return False
            else:
                if not self.should_profile(sampler_key):
                    return False
                
                self._profiling_mode = time.time()
                return True

    def unset_profiling_mode(self):
        with self._profiling_mode_lock:
            self._profiling_mode = None

    def is_profiling_mode(self):
        return self._profiling_mode is not None

    def can_include_profile(self, profile_name) -> bool:
        if self._include_profiles_index is None:
            return True
        return profile_name in self._include_profiles_index

    def emit_span_start(self, span, context):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_span_start(span, context)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_span_stop(self, span, context):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_span_stop(span, context)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_span_read(self, span, context):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_span_read(span, context)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_metric_update(self):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_metric_update()
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def set_tag(self, key: str, value: str, append_uuid: Optional[bool] = False) -> None:
        if not key:
            logger.error('set_tag: key must be provided')
            return

        if value is None:
            self.tags.pop(key, None)
            return

        if len(self.tags) > Tracer.MAX_PROCESS_TAGS:
            logger.error('set_tag: too many tags (>{0})'.format(Tracer.MAX_PROCESS_TAGS))
            return

        if append_uuid:
            if not value:
                value = uuid_sha1(size=12)
            else:
                value = '{0}-{1}'.format(value, uuid_sha1(size=12))

        self.tags[key] = value

    def get_tag(self, key: str) -> Optional[str]:
        return self.tags.get(key, None)

    def remove_tag(self, key: str) -> None:
        self.tags.pop(key, None)

    def set_context_tag(self, key: str, value: str, append_uuid: Optional[bool] = False) -> None:
        if not key:
            logger.error('set_context_tag: key must be provided')
            return

        tags = self.context_tags.get()

        if value is None:
            tags.pop(key, None)
            self.context_tags.set(tags)
            return

        if len(tags) > Tracer.MAX_PROCESS_TAGS:
            logger.error('set_context_tag: too many tags (>{0})'.format(Tracer.MAX_PROCESS_TAGS))
            return

        if append_uuid:
            if not value:
                value = uuid_sha1(size=12)
            else:
                value = '{0}-{1}'.format(value, uuid_sha1(size=12))

        tags[key] = value
        self.context_tags.set(tags)

    def remove_context_tag(self, key: str) -> None:
        tags = self.context_tags.get()
        tags.pop(key, None)
        self.context_tags.set(tags)

    def get_context_tag(self, key: str) -> Optional[str]:
        return self.context_tags.get().get(key, None)

    def trace(
            self, 
            span_name: str,
            tags: Optional[Dict[str, str]] = None,
            include_profiles: Optional[list] = None) -> 'Span':
        return Span(name=span_name, tags=tags, include_profiles=include_profiles)

    def trace_function(
            self, 
            func=None, 
            *,
            span_name: Optional[str] = None,
            tags: Optional[Dict[str, str]] = None,
            include_profiles: Optional[list] = None):
        if func is None:
            return functools.partial(self.trace_function, span_name=span_name, tags=tags, include_profiles=include_profiles)

        if span_name is None:
            span_or_func_name = func.__name__
        else:
            span_or_func_name = span_name

        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def tf_async_wrapper(*args, **kwargs):
                async with self.trace(span_name=span_or_func_name, tags=tags, include_profiles=include_profiles):
                    return await func(*args, **kwargs)
            return tf_async_wrapper
        else:
            @functools.wraps(func)
            def tf_wrapper(*args, **kwargs):
                with self.trace(span_name=span_or_func_name, tags=tags, include_profiles=include_profiles):
                    return func(*args, **kwargs)
            return tf_wrapper

    def report_error(
            self,
            name: str, 
            tags: Optional[Dict[str, str]] = None,
            level: Optional[str] = None,
            message: Optional[str] = None,
            exc_info: Optional[tuple] = None,
            span: Optional[Span] = None) -> None:
        now = int(time.time())

        if not name:
            logger.error('error: name is required')
            return

        if level and level not in self.VALID_ERROR_LEVELS:
            logger.error('error: invalid level "%s", must be one of: %s', level, ', '.join(sorted(self.VALID_ERROR_LEVELS)))
            return

        current_time = time.time()
        if current_time - self._error_counter_reset_time >= 60:
            self._error_counter = 0
            self._error_counter_reset_time = current_time
        
        if self._error_counter >= self.MAX_ERRORS_PER_MINUTE:
            logger.warning('Rate limit exceeded: maximum %d errors per minute', self.MAX_ERRORS_PER_MINUTE)
            return
        
        self._error_counter += 1

        model = client.Error(
            error_id=uuid_sha1(size=12),
            tags=[],
            name=name,
            create_ts=now)

        if span:
            model.linked_span_ids = [span.span_id]

        error_tags = {}
        if self.tags is not None:
            error_tags.update(self.tags)
        if self.context_tags:
            error_tags.update(self.context_tags.get().copy())
        if tags is not None:
            error_tags.update(tags)
        for tag_key, tag_value in error_tags.items():
            model.tags.append(client.Tag(
                key=sanitize_str(tag_key, max_len=50),
                value=sanitize_str(tag_value, max_len=250)))

        if level:
            model.level = level

        # Set message from exc_info if provided and no message given
        if exc_info and not message:
            if exc_info[0] and hasattr(exc_info[0], '__name__'):
                message = str(exc_info[0].__name__)
                message += ': '
            if exc_info[1]:
                message += str(exc_info[1])

        if message:
            model.message = message

        # Extract stack trace from exc_info if provided
        if exc_info:
            frames = traceback.format_tb(exc_info[2])
            if len(frames) > 0:
                model.stack_trace = ''.join(frames)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Uploading error: %s', model)

        self.uploader().upload_error(model)
        self.tick()

    def tick(self, block=False, force=False):
        now = time.time()
        if not force and (now - self.last_tick_ts) < Tracer.TICK_INTERVAL_SEC - 1:
            return
        
        if not self._tick_lock.acquire(blocking=False):
            return

        try:
            def _run_tick():
                try:
                    try:
                        self.emit_metric_update()
                    except Exception as exc:
                        logger.error('Error in metric read event handlers', exc_info=True)

                    if self._metric_store.has_unexported():
                        metrics = self._metric_store.export()
                        for metric in metrics:
                            self.uploader().upload_metric(metric)

                    if self._log_store.has_unexported():
                        entrys = self._log_store.export()
                        for entry in entrys:
                            self.uploader().upload_log_entry(entry)

                    self._uploader.flush()
                except Exception as exc:
                    logger.error('Error in tick execution: %s', exc, exc_info=True)

            self.last_tick_ts = now

            self._tick_run_thread = threading.Thread(target=_run_tick, daemon=True)
            self._tick_run_thread.start()
            if block:
                self._tick_run_thread.join()
        finally:
            self._tick_lock.release()



import logging
import os
import tempfile
import gzip
import shutil
import glob

logger = logging.getLogger('graphsignal')

def create_log_dir():
    log_dir = tempfile.mkdtemp(prefix='graphsignal-')
    logger.debug('Created temporary log directory %s', log_dir)
    return log_dir

def remove_log_dir(log_dir):
    shutil.rmtree(log_dir)
    logger.debug('Removed temporary log directory %s', log_dir)

def find_and_read(log_dir, file_pattern, decompress=True, max_size=None):
    file_paths = glob.glob(os.path.join(log_dir, file_pattern))
    if len(file_paths) == 0:
        logger.debug('Files are not found at %s', os.path.join(log_dir, file_pattern))
        return None

    found_path = file_paths[-1]

    if max_size:
        file_size = os.path.getsize(found_path)
        if file_size > max_size:
            raise Exception('File is too big: {0}'.format(file_size))

    if decompress and found_path.endswith('.gz'):
        last_file = gzip.open(found_path, "rb")
    else:
        last_file = open(found_path, "rb")
    data = last_file.read()
    last_file.close()

    return data


import time
import threading
import logging

import graphsignal
from graphsignal import client
from graphsignal.client.rest import ApiException

logger = logging.getLogger('graphsignal')


class Uploader:
    MAX_BUFFER_SIZE = 10000
    FLUSH_DELAY_SEC = 5

    def __init__(self):
        self._buffer = []
        self._buffer_lock = threading.Lock()
        self._flush_lock = threading.Lock()

    def setup(self):
        pass

    def clear(self):
        with self._buffer_lock:
            self._buffer = []

    def upload_span(self, span):
        self.upload_signal(span)

    def upload_profile(self, profile):
        self.upload_signal(profile)

    def upload_error(self, error):
        self.upload_signal(error)

    def upload_metric(self, metric):
        self.upload_signal(metric)

    def upload_log_entry(self, log_entry):
        self.upload_signal(log_entry)

    def upload_signal(self, signal):
        with self._buffer_lock:
            self._buffer.append(signal)
            if len(self._buffer) > self.MAX_BUFFER_SIZE:
                self._buffer = self._buffer[-self.MAX_BUFFER_SIZE:]

    def flush(self):
        with self._flush_lock:
            with self._buffer_lock:
                if len(self._buffer) == 0:
                    return
                outgoing = self._buffer
                self._buffer = []

            try:
                upload_start = time.time()

                configuration = client.Configuration(
                    host = graphsignal._tracer.api_url,   
                )
                configuration.api_key['APIKeyHeader'] = graphsignal._tracer.api_key
                with client.ApiClient(configuration) as api_client:
                    api_instance = client.DefaultApi(api_client)
    
                    spans = [signal for signal in outgoing if isinstance(signal, client.Span)]
                    if len(spans) > 0:
                        api_instance.upload_spans(spans)

                    profiles = [signal for signal in outgoing if isinstance(signal, client.Profile)]
                    if len(profiles) > 0:
                        api_instance.upload_profiles(profiles)

                    errors = [signal for signal in outgoing if isinstance(signal, client.Error)]
                    if len(errors) > 0:
                        api_instance.upload_errors(errors)

                    metrics = [signal for signal in outgoing if isinstance(signal, client.Metric)]
                    if len(metrics) > 0:
                        api_instance.upload_metrics(metrics)

                    log_entries = [signal for signal in outgoing if isinstance(signal, client.LogEntry)]
                    if len(log_entries) > 0:
                        api_instance.upload_logs(log_entries)

                logger.debug('Upload took %.3f sec', time.time() - upload_start)
            except Exception:
                logger.debug('Failed uploading signals, will retry', exc_info=True)
                with self._buffer_lock:
                    self._buffer[:0] = outgoing



# coding: utf-8

# flake8: noqa
"""
    Graphsignal API

    API for uploading and querying spans, errors, metrics, and logs.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from graphsignal.client.models.attribute import Attribute
from graphsignal.client.models.counter import Counter
from graphsignal.client.models.error import Error
from graphsignal.client.models.error_query_result import ErrorQueryResult
from graphsignal.client.models.http_validation_error import HTTPValidationError
from graphsignal.client.models.histogram import Histogram
from graphsignal.client.models.log_entry import LogEntry
from graphsignal.client.models.metric import Metric
from graphsignal.client.models.metric_type import MetricType
from graphsignal.client.models.profile import Profile
from graphsignal.client.models.profile_query_result import ProfileQueryResult
from graphsignal.client.models.span import Span
from graphsignal.client.models.span_query_result import SpanQueryResult
from graphsignal.client.models.summary import Summary
from graphsignal.client.models.tag import Tag
from graphsignal.client.models.validation_error import ValidationError
from graphsignal.client.models.validation_error_loc_inner import ValidationErrorLocInner



from typing import Dict, Any, Union, Optional, Type
import os
import logging
import atexit

from graphsignal.version import __version__
from graphsignal.env_vars import read_config_param, read_config_tags
from graphsignal.tracer import Tracer
from graphsignal.spans import Span, SpanContext

logger = logging.getLogger('graphsignal')

_tracer = None


def _check_configured():
    global _tracer
    if not _tracer:
        raise ValueError(
            'Tracer not configured, call graphsignal.configure() first')

def configure(
    api_key: Optional[str] = None,
    api_url: Optional[str] = None,
    deployment: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    auto_instrument: Optional[bool] = None,
    samples_per_min: Optional[int] = None,
    profiles_per_min: Optional[int] = None,
    include_profiles: Optional[list] = None,
    debug_mode: Optional[bool] = None
) -> None:
    global _tracer

    if _tracer:
        logger.warning("Tracer already configured")
        return

    api_key = read_config_param("api_key", str, api_key, required=True)
    api_url = read_config_param("api_url", str, api_url)
    tags = read_config_tags(tags)
    auto_instrument = read_config_param("auto_instrument", bool, auto_instrument, default_value=True)
    samples_per_min = read_config_param("samples_per_min", int, samples_per_min)
    profiles_per_min = read_config_param("profiles_per_min", int, profiles_per_min)
    include_profiles = read_config_param("include_profiles", list, include_profiles)
    debug_mode = read_config_param("debug_mode", bool, debug_mode, default_value=False)

    # left for compatibility
    if deployment and isinstance(deployment, str):
        tags['deployment'] = deployment

    _tracer = Tracer(
        api_key=api_key,
        api_url=api_url,
        tags=tags,
        auto_instrument=auto_instrument,
        samples_per_min=samples_per_min,
        profiles_per_min=profiles_per_min,
        include_profiles=include_profiles,
        debug_mode=debug_mode)
    _tracer.setup()

    atexit.register(shutdown)

    logger.debug('Tracer configured')


def set_tag(key: str, value: str) -> None:
    _check_configured()
    _tracer.set_tag(key, value)


def get_tag(key: str) -> Optional[str]:
    _check_configured()
    return _tracer.get_tag(key)


def remove_tag(key: str):
    _check_configured()
    return _tracer.remove_tag(key)


def set_context_tag(key: str, value: str, append_uuid=None) -> None:
    _check_configured()
    _tracer.set_context_tag(key, value, append_uuid=append_uuid)


def get_context_tag(key: str) -> Optional[str]:
    _check_configured()
    return _tracer.get_context_tag(key)


def remove_context_tag(key: str):
    _check_configured()
    return _tracer.remove_context_tag(key)


def trace(
        span_name: str,
        tags: Optional[Dict[str, str]] = None,
        include_profiles: Optional[list] = None) -> 'Span':
    _check_configured()

    return _tracer.trace(span_name=span_name, tags=tags, include_profiles=include_profiles)


def trace_function(
        func=None, 
        *,
        span_name: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        include_profiles: Optional[list] = None) -> Any:
    return _tracer.trace_function(func, span_name=span_name, tags=tags, include_profiles=include_profiles)


def report_error(
        name: str, 
        tags: Optional[Dict[str, str]] = None,
        level: Optional[str] = None,
        message: Optional[str] = None,
        exc_info: Optional[tuple] = None) -> None:
    _check_configured()
    return _tracer.report_error(
        name=name, 
        tags=tags,
        level=level, 
        message=message,
        exc_info=exc_info)


def tick(block=False, force=False) -> None:
    _check_configured()

    _tracer.tick(block=block, force=force)


def shutdown() -> None:
    global _tracer
    if not _tracer:
        return

    atexit.unregister(shutdown)
    _tracer.shutdown()
    _tracer = None

    logger.debug('Tracer shutdown')


__all__ = [
    '__version__',
    'configure',
    'tick',
    'shutdown',
    'trace',
    'trace_function',
    'SpanContext',
    'Span',
    'report_error',
    'set_tag',
    'get_tag',
    'set_context_tag',
    'get_context_tag',
    'remove_context_tag'
]
